/*************************************************************************/
/*
//  Vehicle Routing Problem With GA
//
//
*/
/*************************************************************************/

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <fstream.h>
#include <iostream.h>
#include <iomanip.h>
#include <stdlib.h>
#include <new.h>

#include "TestDefinition.h"

#include "VRPFileReader.h"
#include "VRProblem.h"
#include "VRPSolution.h"
#include "VRPFitnessFunction.h"
#include "VRPInitializer.h"

#include "VRPMutation1PX.h"
#include "VRPMutation1PXG.h"
#include "VRPMutation1PXGPC.h"
#include "VRPMutation2PX.h"
#include "VRPMutation2PXG.h"
#include "VRPMutation2PXGPC.h"

#include "VRPCrossoverSBX.h"
#include "VRPCrossover1PX.h"
#include "VRPCrossover1PXG.h"
#include "VRPCrossover1PXGPC.h"
#include "VRPCrossover2PX.h"
#include "VRPCrossover2PXG.h"
#include "VRPCrossover2PXGPC.h"
#include "VRPCrossoverT1PX.h"
#include "VRPCrossoverT1PXG.h"
#include "VRPCrossoverT2PX.h"
#include "VRPCrossoverDPX.h"

#include "EdgeGraph.h"
#include "VRPPopulation.h"
#include <deffunc.h>
#include <RandomGenerator.h>

#include <Sort_Recombinator.h>
#include <Stat_Recombinator.h>
#include <Generational_Recom.h>

#include <SimilarParentSelector.h>
#include <GeneticAlgorithm.h>

typedef long time_t;
typedef long clock_t;

#define DEBUG 0

/*************************************************************************/
/*
// Memory Allocation Error Handler.
// It will be called when there is no Enough 
// Memory.
*/
/*************************************************************************/


void allocation_handler()
{
   cout << "Out of Memory. Aborting Program...."<< endl;
   exit(1);
}

/*************************************************************************/
/*************************************************************************/

void checkMem(char* mess)
{
   system("ps -l -u htozdemi | grep solveGA > ou");
   ofstream out("mem.trace",ios::app);
   ifstream rin("ou");
   char line[600];
   rin.getline(line,599);
   rin.close();
   if(line!=NULL) out << mess << endl << line << endl;
   out.close();

}// end of checkMem()

/***************************************************************************/
/***************************************************************************/
float f1( float best, float ave, float b, float a, float _val )
{
   float ret = ( _val-ave ) * ( b - a )/(best-ave) + a;
   if( ret<0 ) ret = a;
   else if( ret>1 ) ret = 1;

   return ret;

}// end of f1

/***************************************************************************/
/***************************************************************************/
void replaceCouple( int fst, int snd, Population *pop, ChromosomesBag& chroBag)
{
   int first = fst;
   int second = snd;
   int i,j;
   float *f;
   float *tf;
   float dummy;
   int    size = chroBag.getNumberOfElements();
   f = new float[2+size];
   tf = new float[2+size];


   for(i=0;i<2+size;i++) f[i]=0.0;

   Chromosomes **chromos = new Chromosomes*[2+size];
   Chromosomes *ch = NULL;
   FitnessValue *fVal = NULL;

   chromos[0] = ch = pop->getChromosomesIndex(first);

   if( ch!=NULL ) fVal = ch->getFitnessValue(pop);
   if( fVal!=NULL )
   {
      f[0] = fVal->getLinearValue();
      delete fVal;

   }// end of if

   ch = NULL; fVal = NULL;

   chromos[1] =ch = pop->getChromosomesIndex(second);

   if( ch!=NULL ) fVal = ch->getFitnessValue(pop);
   if( fVal!=NULL )
   {
      f[1] = fVal->getLinearValue();
      delete fVal;

   }// end of if

   if(f[1]>f[0]) {
     i=first; first=second; second=i;
     dummy=f[0];f[0]=f[1];f[1]=dummy;
     ch=chromos[0];chromos[0]=chromos[1];chromos[1]=ch;
   }

   for(i=0;i<size;i++) {

      ch = NULL; fVal = NULL;
      chromos[2+i] = ch = chroBag.getElementIndex(i);

      if( ch!=NULL ) fVal = ch->getFitnessValue(pop);

      if( fVal!=NULL ) { f[2+i] = fVal->getLinearValue(); delete fVal; }

   }// end of for

   for(i=0;i<1+size;i++)
    for(j=i+1;j<2+size;j++)
      if(chromos[i]==chromos[j]) {
	cout << i << " is equal to " << j << endl;
	cout << " fst: " << fst << " snd: " << snd << endl;
	exit(1);
      }

   int bestIndex = pop->getBestIndex();
   ch = pop->getChromosomesIndex(bestIndex);
   fVal = ch->getFitnessValue(pop);
   float pop_best = fVal->getLinearValue();
   delete fVal;

   int besti;
   float tot = 0.0;
   float best = 0.0;
   for(i=0;i<2+size;i++)
   {
     if( best<f[i] ) { best = f[i]; besti=i; }
     tot += f[i];

   }// end of for
   float ave = tot/(2+size);
   cout << " best: " << best << " average: " << ave << endl;
   tot=0;
   for(i=0;i<2+size;i++) 
   { 
      tf[i] = f1(best,ave,0.9,0.5,f[i]);
      cout << "t( "<< f[i] << ") = " << tf[i] << endl;
      tot+=tf[i];
   }

   int betterthansecond=-1;
   if(besti<2) 
     for(i=2;i<2+size;i++)
       if( f[1]<f[i] && (betterthansecond==-1||f[betterthansecond]<f[i]) )
   	   betterthansecond=i; 
   
   float randi = tot * 0.01 * RandomGenerator::generateFloat(0,200);
   float tot1=0;

        if(besti>1&&best>pop_best) i=besti;
   else if(betterthansecond>1) i=betterthansecond;
   else { i=0; tot1=0;
          do { tot1+=tf[i%(size+2)]; i++; } while( tot1<randi );
        }

   int sfirst = i%(2+size);

   i=0;
   randi = tot * 0.01 * RandomGenerator::generateFloat(0,200);
   tot1=0;
   do 
   { if((i%(size+2))!=sfirst) tot1+=tf[i%(size+2)]; 
     i++; 
   } while( tot1<randi );

   int ssecond = i%(2+size);
   if(ssecond==sfirst) ssecond= (sfirst+1)%(2+size);

   if(ssecond==sfirst) {
	cout << " sfirst: " << sfirst << " ssecond: " << ssecond << endl;
	exit(1);
   }// end of if

   /*
      if any child is better than the best in the population,
      then the child will be added to the population.
      else if any child is better then one of the parents(second)
	   then the child will be added to the population.
      else random selection.
   */

   char *mark=new char[2+size];
   for(i=0;i<2+size;i++) mark[i]=1;
   mark[0]=mark[1]=0;

   if( sfirst!=1 ) {  
      ch = chromos[sfirst];
      if(sfirst==0) { 
         VRPChromosomes *Vch = (VRPChromosomes*)ch;
         VRPChromosomes *a = new VRPChromosomes(Vch->getProblem(),Vch->getEdgeGraph());
         *a = *Vch; ch=a;
      }// end of if
      pop->replace(second,ch); mark[1]=1;mark[sfirst]=0; 
   }// end of if

   if( ssecond!=0 && f[ssecond]>f[0] ) { 
      ch = chromos[ssecond];
      if(ssecond==1) { 
         VRPChromosomes *Vch = (VRPChromosomes*)ch;
         VRPChromosomes *a = new VRPChromosomes(Vch->getProblem(),Vch->getEdgeGraph());
         *a = *Vch; ch=a;
      }// end of if
      pop->replace(first,ch);  mark[0]=1;mark[ssecond]=0; 
   }// end of if

   //for(i=0;i<2+size;i++) if(mark[i]) delete chromos[i];

   delete[] mark;
   delete[] f;
   delete[] tf;
   delete[] chromos;

}// end of replaceCouple()

/***************************************************************************/
/***************************************************************************/
void applyLocalSearch(ChromosomesBag* ChroBag) { 
  Chromosomes* currChromosome;
  int currentIndex=0;

  cout << " in applyLocalSearch(ChromosomesBag* ) " << endl;
  while(currentIndex<ChroBag->getNumberOfElements()) {
    if((currChromosome=ChroBag->getElementIndex(currentIndex))!=NULL)
       ((VRPChromosomes*)currChromosome)->applyLocalSearch();
    currentIndex++;
  }// end of while
  cout << " out of applyLocalSearch(ChromosomesBag* ) " << endl;
}// end of applyLocalSearch

/***************************************************************************/
/*
//
// Until the new Population has enough chromosomes in it
//
//   1. Select the parent that undergoes to the Genetic
//      Opertaion.
//
//   2. Flip the coin for Crossover operation. If it
//      succeedes, then apply Crossover.
//
//   3. Flip the coin for Mutation Operation. If it 
//      succeedes, then apply Mutation.
//
*/
/***************************************************************************/

void runOnceOn( CrossOverOperation *CrossOver, MutationOperator *Mutation, 
	        float CrossoverProbability, float MutationProbability, 
	        Population &MyPopulation, float _max, float _min, 
		int& tcross, int& tmutate,int APPLYLOCALSEARCH)
{  int i, k; //, limit = MyPopulation.getSize();
   Chromosomes *Mate;
   ChromosomesBag ChroBag;
   Chromosomes *Mutant;
   Chromosomes *current;
   unsigned opick;
   CrossOverOperation *ops ;
   MutationOperator   *opsM ;
   RandomGenerator Random(0.0,1.0);

   Population SamplePopulation(MyPopulation.getSize());
   time_t start_t;
   time_t stop_t;
   time_t dummy;

   //if(MySampler!=NULL)
   //{
   //    SamplePopulation = MySampler->filterSample(MyPopulation);
   //    *MyNewPop        = MySampler->filterDying(MyPopulation);
   //} 

   SamplePopulation = MyPopulation;

   if(DEBUG) cout << " in GeneticAlgorithm::runEpoch2() " << endl;

   int   currenti, matei;
   FitnessValue *best = SamplePopulation.getBestFitness();
   FitnessValue *ave  = SamplePopulation.getAverageFitness();

   int iterate=1;
   int cycle=0;
   while( iterate )
   {   
       currenti = SamplePopulation.applyTournamentSelection2(_max,_min,best->getLinearValue(),ave->getLinearValue());
       current  = SamplePopulation.getChromosomesIndex(currenti);

       if( CrossoverProbability>0 && ( Random.generate()<CrossoverProbability || (cycle&&iterate) ) )
       {  
	  if(DEBUG) cout << " We need crossover " << endl;

	  matei = currenti;
	  while( matei == currenti )
                 matei = SamplePopulation.applyTournamentSelection2(_max,_min,best->getLinearValue(),ave->getLinearValue());
          Mate  = SamplePopulation.getChromosomesIndex(matei);

	  if(DEBUG) cout << " 2. We need crossover " << endl;

	  if(Mate!=NULL&&CrossOver!=NULL)
          {  
             ops = CrossOver;

	     if(DEBUG) cout << " Before we apply crossover " << endl;
             start_t = time(&dummy);
	     ChroBag = ops->apply(current,Mate,CrossoverProbability);
             stop_t = time(&dummy);
	     if(DEBUG) cout << " After we apply crossover " << endl;
	     tcross += (stop_t-start_t);

	     if( ChroBag.getNumberOfElements()>0 )
	     {  if(APPLYLOCALSEARCH) applyLocalSearch(&ChroBag);
	        replaceCouple(currenti,matei,&MyPopulation,ChroBag);
		iterate = 0;
                current  = SamplePopulation.getChromosomesIndex(currenti);
                best = SamplePopulation.getBestFitness();
                ave  = SamplePopulation.getAverageFitness();

	     }// end of if

	     if(DEBUG) cout << " After we added the result to the Population " << endl;

          } else if(DEBUG) cout << " We could not find any mate " << endl;

       }// end of if

       if( Random.generate()<MutationProbability || (cycle&&iterate) )
       {  
	  if(DEBUG) cout << " We need mutation " << endl;

	  matei = currenti;
	  while( matei == currenti )
                 matei = SamplePopulation.applyTournamentSelection2(_max,_min,best->getLinearValue(),ave->getLinearValue());
          Mate  = SamplePopulation.getChromosomesIndex(matei);

	  if(Mate!=NULL&&Mutation!=NULL)
	  {
             opsM = Mutation;

	     if(DEBUG) cout << " before  mutation.apply()  " << endl;
             start_t = time(&dummy);
	     ChroBag = opsM->apply(current,Mate,MutationProbability);
             stop_t = time(&dummy);
	     if(DEBUG) cout << " after  mutation.apply()  " << endl;
	     tmutate += (stop_t-start_t);

	     if( ChroBag.getNumberOfElements()>0 )
	     {  if(APPLYLOCALSEARCH) applyLocalSearch(&ChroBag);
	        replaceCouple(currenti,matei,&MyPopulation,ChroBag);
		iterate = 0;

	     }// end of if

	  }// end of if

       }// end of if 
       cycle++;

   }// end of for

   if(DEBUG) cout << " End of runEpoch2() " << endl;

}// end of runOnceOn()

/*************************************************************************/
/*************************************************************************/

int decideStat(int& k, int& fre, int i)
{
   int ret = 0;

        if( i<10  ) ret = 1;
   else if( i<20  ) ret = (  (i%2)==0 ) ;
   else if( i<40  ) ret = (  (i%5)==0 ) ; 
   else if( i<80  ) ret = ( (i%10)==0 ) ;
   else if( i<160 ) ret = ( (i%20)==0 ) ;
   else ret = ( (i%50)==0 ) ;

   return ret;

}// end of decideStat()

/*************************************************************************/
/*************************************************************************/

void writeStat(Population* myPop, int ite)
{

   if(DEBUG) cout << " in solveGA2::writeStat() " << endl;

   ofstream out("GASolution.stat",ios::app); 

   out << ite << endl;
   for(int i=0;i<myPop->getSize();i++)
   {
      VRPSolution *curr = (VRPChromosomes *) myPop->getChromosomesIndex(i);
      if(curr!=NULL) 
      {  curr->writeStat(out);
	 FitnessValue *fv = ((VRPChromosomes *)curr)->getFitnessValue(myPop);
	 if(fv!=NULL) { out << " " << *fv ; delete fv; }
	 out << endl;
      }// end of if

   }// end of for

   out << "SofD : " << myPop->getSumOfDissimilarities() 
       << endl << endl;
     
   out.close();
   if(DEBUG) cout << " out of solveGA2::writeStat() " << endl;

}// end of writeStat()

/*************************************************************************/
/*************************************************************************/
void writeMutationType( ostream& out, MutationOprType type) {

   out << " Mutation Oper ..: " ;
   // 1PX, 2PXG, DPX, STD_PX
   switch(  type ) {
      case TestDefinition::R1PX    : out << " VRPMutation1PX() "; break;
      case TestDefinition::R1PXG   : out << " VRPMutation1PXG() "; break;
      case TestDefinition::R1PXGC  : out << " VRPMutation1PXGPC() "; break;
      case TestDefinition::R1PXFG  : out << " VRPMutation1PXFG() "; break;
      case TestDefinition::R1PXFGC : out << " VRPMutation1PXFGPC() "; break;
      case TestDefinition::R2PX    : out << " VRPMutation2PX() "; break;
      case TestDefinition::R2PXG   : out << " VRPMutation2PXG() "; break;
      case TestDefinition::R2PXGC  : out << " VRPMutation2PXGPC() "; break;
      case TestDefinition::R2PXFG  : out << " VRPMutation2PXFG() "; break;
      case TestDefinition::R2PXFGC : out << " VRPMutation2PXFGPC() "; break;
      default:  out << " Unknown "; break;
   }// end of switch
   out << endl << endl;

}// end of writeMutationType()

/*************************************************************************/
/*************************************************************************/
MutationOperator* defineMutationOperator(MutationOprType type) {
   MutationOperator* Jmutation = NULL;
   // 1PX, 2PXG, DPX, STD_PX
   switch( type  ) {
      case TestDefinition::R1PX    : Jmutation = new VRPMutation1PX(); break;
      case TestDefinition::R1PXG   : Jmutation = new VRPMutation1PXG(); break;
      case TestDefinition::R1PXGC  : Jmutation = new VRPMutation1PXGPC(); break;
      case TestDefinition::R1PXFG  : Jmutation = new VRPMutation1PXG(1); break;
      case TestDefinition::R1PXFGC : Jmutation = new VRPMutation1PXGPC(1);break;
      case TestDefinition::R2PX    : Jmutation = new VRPMutation2PX(); break;
      case TestDefinition::R2PXG   : Jmutation = new VRPMutation2PXG(); break;
      case TestDefinition::R2PXGC  : Jmutation = new VRPMutation2PXGPC(); break;
      case TestDefinition::R2PXFG  : Jmutation = new VRPMutation2PXG(1); break;
      case TestDefinition::R2PXFGC : Jmutation = new VRPMutation2PXGPC(1);break;
      default: Jmutation = new VRPMutation1PX(); break;
   }// end of switch
   return Jmutation;
}// end of defineMutationOperator()

/*************************************************************************/
/*************************************************************************/
void writeMutationType(ostream& out, TestDefinition* tst)
{ writeMutationType(out,tst->getMutationOprType()); }

/*************************************************************************/
/*************************************************************************/
void writeCrossoverType( ostream& out, CrossoverOprType type) {

   out << " Crosover Oper ..: " ;
   // STD_PX, CR1PX, CR1PXG, TIME_1PX, TIME_2PX, DPX, STD_PX
   switch(  type ) {
      case TestDefinition::STD_PX  : out << " VRPCrossoverSBX() "; break;
      case TestDefinition::CR1PX   : out << " VRPCrossover1PX() "; break;
      case TestDefinition::CR1PXG  : out << " VRPCrossover1PXG() "; break;
      case TestDefinition::CR1PXGC : out << " VRPCrossover1PXGPC() "; break;
      case TestDefinition::CR1PXFG : out << " VRPCrossover1PXFG() "; break;
      case TestDefinition::CR1PXFGC: out << " VRPCrossover1PXFGPC() "; break;
      case TestDefinition::CR2PX   : out << " VRPCrossover2PX() "; break;
      case TestDefinition::CR2PXG  : out << " VRPCrossover2PXG() "; break;
      case TestDefinition::CR2PXGC : out << " VRPCrossover2PXGPC() "; break;
      case TestDefinition::CR2PXFG : out << " VRPCrossover2PXFG() "; break;
      case TestDefinition::CR2PXFGC: out << " VRPCrossover2PXFGPC() "; break;
      case TestDefinition::T1PX    : out << " VRPCrossoverT1PX() "; break;
      case TestDefinition::T1PXG   : out << " VRPCrossoverT1PXG() "; break;
      case TestDefinition::T1PXGC  : out << " VRPCrossoverT1PXGC() "; break;
      case TestDefinition::T2PX    : out << " VRPCrossoverT2PX() "; break;
      case TestDefinition::DPX     : out << " VRPCrossoverDPX() "; break;
      default: out << " Unknown "; break;
   }// end of switch
   out << endl << endl;

}// end of writeCrossoverType()

/*************************************************************************/
/*************************************************************************/
CrossOverOperation* defineCrossOverOperation(CrossoverOprType type, int noc) {
   CrossOverOperation* Jcross = NULL;
   // STD_PX, CR1PX, CR1PXG, TIME_1PX, TIME_2PX, DPX, STD_PX
   switch( type ) {
      case TestDefinition::STD_PX  : Jcross = new VRPCrossoverSBX(noc); break;
      case TestDefinition::CR1PX   : Jcross = new VRPCrossover1PX(); break;
      case TestDefinition::CR1PXG  : Jcross = new VRPCrossover1PXG(); break;
      case TestDefinition::CR1PXGC : Jcross = new VRPCrossover1PXGPC(); break;
      case TestDefinition::CR1PXFG : Jcross = new VRPCrossover1PXG(1); break;
      case TestDefinition::CR1PXFGC: Jcross = new VRPCrossover1PXGPC(1); break;
      case TestDefinition::CR2PX   : Jcross = new VRPCrossover2PX(); break;
      case TestDefinition::CR2PXG  : Jcross = new VRPCrossover2PXG(); break;
      case TestDefinition::CR2PXGC : Jcross = new VRPCrossover2PXGPC(); break;
      case TestDefinition::CR2PXFG : Jcross = new VRPCrossover2PXG(1); break;
      case TestDefinition::CR2PXFGC: Jcross = new VRPCrossover2PXGPC(1); break;
      case TestDefinition::T1PX    : Jcross = new VRPCrossoverT1PX(); break;
      case TestDefinition::T1PXG   : Jcross = new VRPCrossoverT1PXG(noc,1); break;
      case TestDefinition::T1PXGC  : Jcross = new VRPCrossoverT1PXG(noc,2); break;
      case TestDefinition::T2PX    : Jcross = new VRPCrossoverT2PX(); break;
      case TestDefinition::DPX     : Jcross = new VRPCrossoverDPX(); break;
      default: Jcross = new VRPCrossoverSBX(noc); break;
   }// end of switch
   return Jcross;
}// end of define

/*************************************************************************/
/*************************************************************************/
void writeCrossoverType(ostream& out, TestDefinition* tst)
{ writeCrossoverType(out,tst->getCrossoverOprType()); }

/*************************************************************************/
/*************************************************************************/
void writeFitnessValues(int i, Population& myPop, int PopulationSize , 
			unsigned *sIndex, char *msg)
{
   int m;
   FitnessValue *fv;
   ofstream oStatT("GASolution.sort",ios::app); 
   oStatT << "  After " << i << " epoch PS " << PopulationSize ;
   if(msg!=NULL) oStatT << msg ; 
   oStatT << endl ;
   for(m=0;m<PopulationSize;m++)
   {
      Chromosomes* chr = myPop.getChromosomesIndex(sIndex[m]);
      if(chr!=NULL && (fv=chr->getFitnessValue(&myPop))!=NULL) 
      {  oStatT << "[ " << setw(2) <<m << " - " << setw(2) << sIndex[m] <<"] : " << *fv << endl;
         delete fv;
       }// end of if
        
   }// end of for
   oStatT.close();

}// end of writeFitnessValues()

/*************************************************************************/
/*************************************************************************/
void writeFitnessValues(int i, Population& myPop, int PopulationSize , 
			unsigned *sIndex, char* mark, char *msg)
{
   int m;
   FitnessValue *fv;
   ofstream oStatT("GASolution.sort",ios::app); 
   oStatT << "  After " << i << " epoch PS " << PopulationSize ;
   if(msg!=NULL) oStatT << msg ; 
   oStatT << endl ;
   for(m=0;m<PopulationSize;m++)
   {
      Chromosomes* chr = myPop.getChromosomesIndex(sIndex[m]);
      if(chr!=NULL && (fv=chr->getFitnessValue(&myPop))!=NULL) 
      {  oStatT << "[ " << setw(2) << m << " - " 
			<< setw(2) << sIndex[m] 
			<< " (" << (int)mark[m] << ")" 
			<<" ] : " << *fv << endl;
         delete fv;
       }// end of if
        
   }// end of for
   oStatT.close();

}// end of writeFitnessValues()

/*************************************************************************/
/*************************************************************************/
void writePopulation( ostream& oStat1, int i, int MaxGenNum, Population* myPop, 
		      long bias, time_t stop_t, time_t start_t, int tcross, 
		      int tmutate, time_t stattime, int numOfRestart, 
		      long totRestartTime, TestDefinition *tst, 
		      int APPLYLOCALSEARCH)
{

   oStat1 << " After " << i << "/" << MaxGenNum << " epoch " << endl 
          << " Elapsed Time " << bias+(stop_t-start_t) << " : "
	  << (double)(bias+(stop_t-start_t))/3600.0 << endl
          << " Elapsed Time for Initialize " << bias << " : "
	  << (double)bias/3600.0 << endl
          << " Elapsed Time for Crossover " << tcross << " : "
	  << (double)tcross/3600.0 << endl
          << " Elapsed Time for Mutation " << tmutate << " : "
	  << (double)tmutate/3600.0 << endl
          << " Elapsed Time for StatWriting " << stattime << endl
          << " Numner of restarts " << numOfRestart << " " 
	  << " Total time spent for restarts "<< totRestartTime << endl;

   oStat1 << " Problem Data file name : " << tst->getDataFileName() << endl;
   oStat1 << " Sum of Dissimilarities() " 
	  << myPop->getSumOfDissimilarities() << endl;

   writeCrossoverType(oStat1,tst);
   writeMutationType(oStat1,tst);
   oStat1 << " Local Search is " << (APPLYLOCALSEARCH?"":"*NOT*") 
	  << " applied. "<< endl << endl;

   oStat1 << *myPop << endl;

}// end of writePopulation()

/*************************************************************************/
/*************************************************************************/
void writePopulation(char *fName, int oMode, int i, int MaxGenNum, 
		     Population* myPop, long bias, time_t stop_t, 
		     time_t start_t, int tcross, int tmutate, time_t stattime, 
		     int numOfRestart, long totRestartTime, 
		     TestDefinition *tst, int APPLYLOCALSEARCH)
{

   ofstream oStat1(fName,oMode);

   writePopulation(oStat1,i,MaxGenNum,myPop,bias,stop_t,start_t,tcross,tmutate,
		   stattime,numOfRestart,totRestartTime,tst,APPLYLOCALSEARCH);


   oStat1.close();

}// end of writePopulation()

/*************************************************************************/
/*************************************************************************/

#define POP_SIZE  15
#define MAX_GEN   100

void solvewithGASTDptr(VRProblem *VRPInstance, TestDefinition& tst, int APPLYLOCALSEARCH)
{
   VRPFitnessFunction   fitFunc;
   VRPInitializer       Initme(VRPInstance);
   MutationOperator     *Jmutation=NULL;
   CrossOverOperation   *Jcross= NULL; 

   VRPPopulation         myPop(tst.getPopulationSize());    

   time_t start_t;
   time_t stop_t;
   time_t dummy;
   long   bias = 0;
   long   totRestartTime = 0;

   time_t dstart_t;
   time_t dstop_t;
   time_t stattime = 0;

   int tcross=0;
   int tmutate=0;
   char *mark;

   checkMem(" before Initme.apply(myPop) ");
   cout << " before init() " << endl;

   start_t = time(&dummy);
   Initme.apply(myPop);
   stop_t = time(&dummy);
   bias = stop_t - start_t;
   cout << " after init() " << endl;
   checkMem(" after Initme.apply(myPop) ");

   Jmutation=defineMutationOperator(tst.getMutationOprType());
   Jcross=defineCrossOverOperation(tst.getCrossoverOprType(),VRPInstance->getNumberOfCustomers());
   mark = new char[tst.getPopulationSize()];

   //Sort_Recombinator     SRecombinator;
   Stat_Recombinator     SRecombinator;
   SimilarParentSelector SimilarSelect;
    
   GeneticAlgorithm *mySolver = new GeneticAlgorithm(&fitFunc,Jcross,Jmutation,&SimilarSelect,&SRecombinator,&Initme);

   int MaxGenNum = tst.getNumberOfGenerations();
   int PopulationSize = tst.getPopulationSize();
   float MutationProbability = tst.getCrossProbability();
   float CrossoverProbability = tst.getMutationProbability();
   int TestTimes = 1;
   int TestType = 0;

   mySolver->SetMaxGeneration(MaxGenNum);
   mySolver->SetPopulationSize(PopulationSize);
   mySolver->SetMutationProb(MutationProbability);
   mySolver->SetCrossProb(CrossoverProbability);

   if(TestType>0&&TestType<4) mySolver->setTestType(TestType);

   if(DEBUG) cout << " before run() " << endl;

   if(DEBUG) cout << " Current population " << myPop << endl << "0000" << endl;
   Population *nPop ;

   if(DEBUG) {
   ofstream oStat2("GASolution.ini"); 
   oStat2 << " Initial Population " << endl << myPop << endl;
   oStat2.close();
   }

   ofstream oStat3("GASolution.stat"); 
   oStat3 << MaxGenNum << " " << PopulationSize <<" " << endl;
   oStat3.close();

   ofstream outT("replace.trace",ios::app);
   outT.close();

   char fileName[30];
   sprintf(fileName,"GAStatisticsT.out");

   ofstream oStat(fileName); oStat.close();

   int numOfRestart=0;
   int k=0,fre=1;

   start_t = time(&dummy);
   for(int i=0;i<MaxGenNum;i++)
   {
      cout << " gen: " << i << "/" << MaxGenNum << endl;
      checkMem(" before generation ");
      if(DEBUG) cout << endl << endl << " *** Before writeStat()  " << i << endl;
      dstart_t = time(&dummy);
      if(decideStat(k,fre,i)) writeStat(&myPop,i);
      dstop_t = time(&dummy);
      stattime += (dstop_t-dstart_t) ;

      if(DEBUG) cout << endl << endl << " GENERATION # " << i << endl << endl;

      runOnceOn( Jcross,Jmutation,CrossoverProbability,MutationProbability,
		 myPop,0.9,0.5,tcross,tmutate,APPLYLOCALSEARCH);

      myPop.writeStatistics(i,fileName);

      if(decideStat(k,fre,i))
        writePopulation("GASolution.out",ios::out,i,MaxGenNum,&myPop,bias,
			stop_t, start_t,tcross,tmutate,stattime,numOfRestart,
			totRestartTime,&tst,APPLYLOCALSEARCH);

      checkMem(" after generation ");

      if( decideStat(k,fre,i) && myPop.getSumOfDissimilarities() < 0.1)
      {
	 numOfRestart++;
         dstart_t = time(&dummy);
	 int limit, m, k, nOfreplace = maximum(1,0.7*PopulationSize);
	 unsigned* sIndex = myPop.getSortIndex();

         writeFitnessValues(i,myPop,PopulationSize,sIndex," before restart");

	 limit=PopulationSize-nOfreplace;

	 // If top chromosomes are close to each other then apply
	 // randomization
	 //for( k=0, m=0; m<PopulationSize && k<(PopulationSize-nOfreplace);m++)
	 //{
	 //  int m1,nok=1;
         //  Chromosomes* chr1 = myPop.getChromosomesIndex(sIndex[m]);
	 //  if( chr1!=NULL )
	 //   for(m1=m+1;nok&&m1<limit;m1++)
	 //   {
	 //     Chromosomes* chr2 = myPop.getChromosomesIndex(sIndex[m1]);
	 //     if( chr2!=NULL && chr1->HammingDistance(chr2)<0.0001) 
	 //     {  
		// chr2->randomInitialize(); 
	      //   nok=0; 
	//	 limit++;
	 //     } else k++;
	 //   }// end of for
	 //}// end of for

	 //
	 //for(k=0;k<m;k++)
	 //{
	 //   Chromosomes* chr = myPop.getChromosomesIndex(sIndex[k]);
	 //   if(chr!=NULL) ((AirChromosomes*)chr)->applyLocalSearch2();
   	 //}// end of for

	 // initialize mark
	 for(k=0;k<PopulationSize;k++) mark[k]=0;
	 cout << " mark is initialized " << endl;
         int prv,nxt;
	 FitnessValue* fprv;
	 FitnessValue* fnxt;
	 prv=0;
	 fprv=myPop.getFitnessValueIndex(sIndex[prv]);
	 while(prv<PopulationSize-1) {
	   nxt=prv+1;
	   fnxt=myPop.getFitnessValueIndex(sIndex[nxt]);
	   //while(nxt<PopulationSize&&fnxt->getLinearValue()==fprv->getLinearValue()) {
	   while(nxt<PopulationSize&&*fnxt==*fprv) {
	     delete fnxt;
	     mark[nxt]=1;
	     nxt++;
	     fnxt=(nxt<PopulationSize?myPop.getFitnessValueIndex(sIndex[nxt]):NULL);
	   }// end of while
	   delete fprv;
	   fprv=fnxt;
	   prv=nxt;
	 }// end of while()
	 if(fprv!=NULL) delete fprv;

         writeFitnessValues(i,myPop,PopulationSize,sIndex,mark," for mark");
	 Chromosomes* chr = NULL;
         int alwaysRestart=0; 
	 prv=0;nxt=0;
	 while(prv<PopulationSize) {

	   if( ( mark[prv] || alwaysRestart ) &&
	       ( (chr=myPop.getChromosomesIndex(sIndex[prv]))!=NULL ) ) 
	        chr->randomInitialize();
           else nxt++;
	   if(nxt>limit) alwaysRestart=1;
	   prv++;
	 }// end of while

	 /*
	 Chromosomes* chr = NULL;
	 for(m=limit;m<PopulationSize;m++)
	   if( (chr=myPop.getChromosomesIndex(sIndex[m]))!=NULL) { 
	      chr->randomInitialize();
	      //myPop.replace(sIndex[m],chr);
	   }// end of if
         */

	 //for(m=PopulationSize-nOfreplace-1;m<PopulationSize;m++)
	 //{
	 //   Chromosomes* chr = myPop.getChromosomesIndex(sIndex[m]);
	 //   if(chr!=NULL) 
	 //   { 
	 //      chr->randomInitialize();
	 //      myPop.replace(sIndex[m],chr);
	 //   }// end of if
	 //}// end of for

         writeFitnessValues(i,myPop,PopulationSize,sIndex," after restart");

	 delete[] sIndex;
         dstop_t = time(&dummy);
	 totRestartTime += (dstop_t-dstart_t);
	    
      }// end of if

   }// end of for

   stop_t = time(&dummy);

   writeStat(&myPop,i);

   if(DEBUG) cout << " After one epoch " << myPop << endl;

   writePopulation("GASolution.out",ios::out,i,MaxGenNum,&myPop,bias,stop_t,start_t,
		   tcross, tmutate,stattime,numOfRestart,totRestartTime,&tst,APPLYLOCALSEARCH);

   //
   unsigned *srtIndex = myPop.getSortIndex();
   Chromosomes *best = myPop.getChromosomesIndex(srtIndex[0]);
   ofstream oStatZ("GASolution.best");
   ((VRPChromosomes*)best)->writeCoded(oStatZ);
   oStatZ.close();
   delete[] srtIndex;

   writePopulation( cout,i,MaxGenNum,&myPop,bias,stop_t,start_t,tcross,tmutate,
		    stattime,numOfRestart,totRestartTime,&tst,APPLYLOCALSEARCH);

   delete[] mark;
   return;

}// end of solvewithGASTDptr()

/*************************************************************************/
/*************************************************************************/

void selectMates( Population &MyPopulation, Chromosomes** first, 
		  Chromosomes** second, float _max, float _min)
{  int   currenti, matei;
   Population SamplePopulation(MyPopulation.getSize());
   time_t start_t;
   time_t stop_t;
   time_t dummy;
   Chromosomes* bb;

   SamplePopulation = MyPopulation;
   if(DEBUG) cout << " in GeneticAlgorithm::selectMates() " << endl;

   FitnessValue *best = SamplePopulation.getBestFitness();
   FitnessValue *ave  = SamplePopulation.getAverageFitness();
   currenti = SamplePopulation.applyTournamentSelection2(_max,_min,best->getLinearValue(),ave->getLinearValue());
   *first  = SamplePopulation.getChromosomesIndex(currenti);

   matei = currenti;
   while( matei == currenti )
          matei = SamplePopulation.applyTournamentSelection2(_max,_min,best->getLinearValue(),ave->getLinearValue());
   *second = SamplePopulation.getChromosomesIndex(matei);
   FitnessValue* ff = (*first)->getFitnessValue();
   FitnessValue* fs = (*second)->getFitnessValue();
   if(*ff<*fs) { bb = *first; *first=*second; *second=bb; }
   delete ff;
   delete fs;

}// end of selectMates()

/*************************************************************************/
/*************************************************************************/

ChromosomesBag applyMUT( MutationOperator *Mutation, Chromosomes *current, 
			 Chromosomes *Mate, time_t& tmutate)
{  ChromosomesBag ChroBag;

   time_t start_t;
   time_t stop_t;
   time_t dummy;

   if( current!=NULL && Mate!=NULL && Mutation!=NULL ) {
       MutationOperator* opsM = Mutation;
       if(DEBUG) cout << " before  mutation.apply()  " << endl;
       start_t = time(&dummy);
       ChroBag = opsM->apply(current,Mate,0.5);
       stop_t = time(&dummy);
       if(DEBUG) cout << " after  mutation.apply()  " << endl;
       tmutate = (stop_t-start_t);
   }// end of if 

   if(DEBUG) cout << " End of applyMUT() " << endl;

   return ChroBag;

}// end of applyMUT()

/*************************************************************************/
/*************************************************************************/

ChromosomesBag applyCPX( CrossOverOperation *CrossOver, Chromosomes *current, 
		         Chromosomes *Mate, time_t& tcross)
{  ChromosomesBag ChroBag;

   time_t start_t;
   time_t stop_t;
   time_t dummy;

   if(DEBUG) cout << " 2. We need crossover " << endl;

   if( current!=NULL && Mate!=NULL && CrossOver!=NULL ) {  
       CrossOverOperation* ops = CrossOver;
       if(DEBUG) cout << " Before we apply crossover " << endl;
       start_t = time(&dummy);
       ChroBag = ops->apply(current,Mate,0.5);
       stop_t = time(&dummy);
       if(DEBUG) cout << " After we apply crossover " << endl;
       tcross = (stop_t-start_t);
   } else if(DEBUG) cout << " We could not find any mate " << endl;

   if(DEBUG) cout << " End of applyCPX() " << endl;

   return ChroBag;

}// end of applyCPX()

/*************************************************************************/
/*************************************************************************/

void write( char* fName, VRPChromosomes *first, VRPChromosomes *second, 
	    ChromosomesBag* ChroBag)
{
  if(DEBUG) cout << " in solveGA::write() " << endl;

  ofstream out(fName,ios::app); 

  out << " Cost[parent #1]= " << first->getCost() 
       << " Cost[parent #2]= " << second->getCost() << endl;

  Chromosomes* currChromosome;
  int currentIndex=0;

  while(currentIndex<ChroBag->getNumberOfElements()) {
    if((currChromosome=ChroBag->getElementIndex(currentIndex))!=NULL) {
      out << " Cost[Child  #"<< currentIndex << "]= " 
          << ((VRPChromosomes*)currChromosome)->getCost() << " ";
      if(((VRPChromosomes*)currChromosome)->getCost()<first->getCost()) out<<"+";
      else out<<"-";
      if(((VRPChromosomes*)currChromosome)->getCost()<second->getCost()) out<<"+";
      else out<<"-";
      out << endl;
    }// end of if
    currentIndex++;
  }// end of while

  out.close();
  if(DEBUG) cout << " out of solveGA::write() " << endl;

}// end of write()

/*************************************************************************/
/*************************************************************************/

void reportPerformance( char* fName, time_t* opr_perf, int limit) {

  if(DEBUG) cout << " in solveGA::reportPerformance() " << endl;

  time_t total = 0;
  time_t ave   = 0;
  time_t min   = (time_t)opr_perf[0];
  time_t max   = (time_t)opr_perf[0];
  int i;
  ofstream out(fName,ios::app); 

  for(i=0;i<limit;i++) {
    total+=opr_perf[i];
    if(opr_perf[i]>max) max=opr_perf[i];
    else if(opr_perf[i]<min) min=opr_perf[i];
  }// end of for

  if(limit>0) ave=1.0*total/limit;

  out << " Minimum time ..................: " << min << endl;
  out << " Maximum time ..................: " << max << endl;
  out << " Average time ..................: " << ave << endl;

  out.close();

  if(DEBUG) cout << " out of solveGA::reportPerformance() " << endl;

}// end of writePerformance()

/*************************************************************************/
/*************************************************************************/

void writePerformance( char* fName, int numofchildren,
			 int numofchildrenbetterthanfirst, 
			 int numofchildrenbetterthansecond,
			 int numofchildrenbetterthanboth) {
  if(DEBUG) cout << " in solveGA::writePerformance() " << endl;

  ofstream out(fName,ios::app); 

  out << " # of children  (NC) ...........: " << numofchildren << endl;
  out << " # of better than first (NBF) ..: " 
      << numofchildrenbetterthanfirst << endl;
  out << " # of better than second (NBS) .: " 
      << numofchildrenbetterthansecond << endl;
  out << " # of better than both (NBB) ...: " 
      << numofchildrenbetterthanboth << endl;
  out << " NBF/NC ........................: " 
      << 1.0*numofchildrenbetterthanfirst/numofchildren << endl;
  out << " NBS/NC ........................: " 
      << 1.0*numofchildrenbetterthansecond/numofchildren << endl;
  out << " NBB/NC ........................: " 
      << 1.0*numofchildrenbetterthanboth/numofchildren << endl;
  out.close();

  if(DEBUG) cout << " out of solveGA::writePerformance() " << endl;

}// end of writePerformance()

/*************************************************************************/
/*************************************************************************/
void collectPerformance( VRPChromosomes *first, VRPChromosomes *second, 
	                 ChromosomesBag* ChroBag, int& numofchildren,
			 int& numofchildrenbetterthanfirst, 
			 int& numofchildrenbetterthansecond,
			 int& numofchildrenbetterthanboth) {
  Chromosomes* currChromosome;
  int currentIndex=0;
  numofchildren+=ChroBag->getNumberOfElements();

  while(currentIndex<ChroBag->getNumberOfElements()) {
    if((currChromosome=ChroBag->getElementIndex(currentIndex))!=NULL) {

      if(((VRPChromosomes*)currChromosome)->getCost()<first->getCost())
        numofchildrenbetterthanfirst++;
      if(((VRPChromosomes*)currChromosome)->getCost()<second->getCost())
        numofchildrenbetterthansecond++;
      if( ((VRPChromosomes*)currChromosome)->getCost()<first->getCost()&&
          ((VRPChromosomes*)currChromosome)->getCost()<second->getCost())
        numofchildrenbetterthanboth++;
    }// end of if
    currentIndex++;
  }// end of while
}// end of collectPerformance()

/*************************************************************************/
/*************************************************************************/

void applyOperator( VRProblem *VRPInstance, TestDefinition& tst, 
		    int isMutation, int refresh)
{
   VRPFitnessFunction   fitFunc;
   VRPInitializer       Initme(VRPInstance);
   MutationOperator*   Jmutation=defineMutationOperator(tst.getMutationOprType());
   CrossOverOperation* Jcross=defineCrossOverOperation(tst.getCrossoverOprType(),
					  VRPInstance->getNumberOfCustomers());


   VRPPopulation         myPop(tst.getPopulationSize());    

   time_t start_t;
   time_t stop_t;
   time_t dummy;
   time_t opr_app;
   long   bias = 0;
   long   totRestartTime = 0;

   time_t dstart_t;
   time_t dstop_t;
   time_t stattime = 0;

   int tcross=0;
   int tmutate=0;

   checkMem(" before Initme.apply(myPop) ");
   cout << " before init() " << endl;

   start_t = time(&dummy);
   Initme.apply(myPop);
   stop_t = time(&dummy);
   bias = stop_t - start_t;
   cout << " after init() " << endl;
   checkMem(" after Initme.apply(myPop) ");

   //Sort_Recombinator     SRecombinator;
   Stat_Recombinator     SRecombinator;
   SimilarParentSelector SimilarSelect;
    
   GeneticAlgorithm *mySolver = new GeneticAlgorithm(&fitFunc,Jcross,Jmutation,&SimilarSelect,&SRecombinator,&Initme);

   int MaxGenNum = tst.getNumberOfGenerations();
   int PopulationSize = tst.getPopulationSize();
   float MutationProbability = tst.getCrossProbability();
   float CrossoverProbability = tst.getMutationProbability();
   int TestTimes = 1;
   int TestType = 0;

   mySolver->SetMaxGeneration(MaxGenNum);
   mySolver->SetPopulationSize(PopulationSize);
   mySolver->SetMutationProb(MutationProbability);
   mySolver->SetCrossProb(CrossoverProbability);

   if(TestType>0&&TestType<4) mySolver->setTestType(TestType);

   if(DEBUG) cout << " before run() " << endl;

   if(DEBUG) cout << " Current population " << myPop << endl << "0000" << endl;
   Population *nPop ;

   if(DEBUG) {
    ofstream oStat2("GASolution.ini"); 
    oStat2 << " Initial Population " << endl << myPop << endl;
    oStat2.close();
   }// end of if

   ofstream oStat3("Operation.stat"); 
   oStat3 << " # of Trials ....: " << MaxGenNum << endl;
   oStat3 << " PopulationSize .: "  << PopulationSize << endl;
   if(isMutation) writeMutationType(oStat3,tst.getMutationOprType());
   else writeCrossoverType(oStat3,tst.getCrossoverOprType());
   oStat3.close();

   ofstream outT("replace.trace",ios::app);
   outT.close();

   char fileName[30];

   int numOfRestart=0;
   int k=0,fre=1;
   Chromosomes* first;
   Chromosomes* second;
   Chromosomes* tmp;
   selectMates(myPop,&first,&second,0.9,0.5);
   ChromosomesBag ChroBag(2);
   if(((VRPChromosomes*)second)->getCost()<((VRPChromosomes*)first)->getCost())
   { tmp=first; first=second; second=tmp; }
   int numofchildren=0;
   int numofchildrenbetterthanfirst=0;
   int numofchildrenbetterthansecond=0;
   int numofchildrenbetterthanboth=0;
   time_t *opr_perf = new time_t[MaxGenNum];

   start_t = time(&dummy);
   for(int i=0;i<MaxGenNum;i++) {
      cout << " Test #: " << i << "/" << MaxGenNum << endl;
      checkMem(" before generation ");

      if(isMutation) ChroBag=applyMUT(Jmutation,first,second,opr_app);
      else ChroBag=applyCPX(Jcross,first,second,opr_app);
      opr_perf[i]=(opr_app>0?opr_app:0);

      write("Operation.stat",(VRPChromosomes*)first,(VRPChromosomes*)second,&ChroBag);
      collectPerformance((VRPChromosomes*)first,(VRPChromosomes*)second,
			 &ChroBag,numofchildren,numofchildrenbetterthanfirst,
			 numofchildrenbetterthansecond,numofchildrenbetterthanboth);

      if(refresh) {
        selectMates(myPop,&first,&second,0.9,0.5);
        if(((VRPChromosomes*)second)->getCost()<((VRPChromosomes*)first)->getCost())
        { tmp=first; first=second; second=tmp; }
      }// end of if 

   }// end of for
   stop_t = time(&dummy);


   writePerformance("Operation.stat",numofchildren,numofchildrenbetterthanfirst,
		    numofchildrenbetterthansecond,numofchildrenbetterthanboth);
   reportPerformance("Operation.stat",opr_perf,MaxGenNum);
   delete[] opr_perf;
   return;

}// end of applyOperator()

/**
* fileName gives the TestDefinition file
*/
void solveGA(char *fileName, char* choice)
{  
   int i=0; 

   ofstream outK("mem.trace"); outK.close();

   // Set Allocation Exception Handler
   set_new_handler(allocation_handler);
   srand(150);

   /*
   if(fileName!=NULL) cout << " fileName: " << fileName << endl;
   if(choice!=NULL) cout << " choice: " << choice << endl;
   if(choice!=NULL&&choice[0]=='o') 
      cout << "applyOperator(&Prob,tst,"
	   <<(int)(choice[1]=='m') << ","
	   <<(int)(choice[2]=='r')
   	   <<");"<< endl;
   else cout << " solvewithGASTDptr(&Prob,tst); " << endl;
   if(i<1) exit(1);
   */
   ifstream test1(fileName);
   if(!test1.good()) {
     cout << fileName << " does not exist! " << endl;
     exit(1);
   } // end of if
   test1.close();

   TestDefinition tst ;
   tst = tst.readTestDefinition(fileName);

   // Read the data file
   VRPFileReader Reader(tst.getDataFileName());  
   VRProblem Prob ;
   Reader.read(Prob);

   // Update Depots 
   gVector<customerId> depots = tst.getDepots();
   customerId *id;
   for(i=0;i<depots.getSize();i++) 
     if((id=depots.getNthElement(i))!=NULL) Prob.addDepot(*id);

   // Print out the current problem
   if(DEBUG) cout << Prob ;

   // Set the TestDefinition Object
   Prob.setTestDefinition(tst);

   // Build EdgeGraph
   EdgeGraph eg(&Prob,Prob.getNumberOfCustomers());

   eg.buildGraphFor(&Prob);
   Prob.setEdgeGraph(&eg); 

   //Prob.buildPreSolutions();

   // Print out EdgeGraph
   if(DEBUG) cout << eg << endl;

   // Solve with GA
   cout << "\n with GA " << endl;
   if(choice!=NULL&&choice[0]=='o') 
	applyOperator(&Prob,tst,(choice[1]=='m'),(choice[2]=='r'));
   else solvewithGASTDptr(&Prob,tst,(choice!=NULL&&choice[0]=='l'));
   cout << "\n end GA " << endl;

}// end of solveGA()

void printArgs(int argc, char *argv[]) {
  cout << " argc " << argc << endl;
  for(int i=0;i<argc;i++)
   cout << " argv[ " << i << "] : " << argv[i] << endl;
}

/** */
main(int argc, char *argv[])
{
   printArgs(argc,argv);
   if(argc>1) solveGA(argv[1],argv[2]);
   else
    {
       cout << " usage :  solveGA <testdefinitionfilename> " << endl;
       cout << " test definition file name describes the current test " << endl;
       cout << " parameters. " << endl << endl << endl;
    }

}// end of main

